objectdef bindbuilder
{
	variable string Name
	variable string Hotkey

	variable string PressCmd
	variable string ReleaseCmd

	variable bool Hold
	variable bool UseFTLModifiers
	variable string SendNextClick

	method AddPressCommand(string value)
	{
		if ${PressCmd.Length}
			PressCmd:Concat[";${value.Escape}"]
		else
			PressCmd:Set["${value.Escape}"]
	}
	method AddReleaseCommand(string value)
	{
		if ${ReleaseCmd.Length}
			ReleaseCmd:Concat[";${value.Escape}"]
		else
			ReleaseCmd:Set["${value.Escape}"]
	}

	member FTLModifiers()
	{
		if ${UseFTLModifiers}
			return ${KeyMapper.FTLModifiers.Escape}
		return
	}

	member ResolveTarget(string name)
	{
		if ${name.Equal[self]}
			return
		if ${name.Equal[${Int["${name.Escape}"]}]}
			return "is${name.Escape}"
		return "${name.Escape}"
	}

	method AddKeypress(string name, string key)
	{
		if ${name.Equal[self]}
		{
			if ${Hold}
			{
				This:AddPressCommand["press -hold \"${key.Escape}\""]
				This:AddReleaseCommand["press -release \"${key.Escape}\""]
			}
			else
			{
				This:AddReleaseCommand["press -nomodifiers \"${key.Escape}\""]
			}
		}
		else
		{

			if ${Hold}
			{
				This:AddPressCommand["relay \"${This.ResolveTarget["${name.Escape}"]}\" press -hold \"${This.FTLModifiers}${key.Escape}\""]
				This:AddReleaseCommand["relay \"${This.ResolveTarget["${name.Escape}"]}\" press -release \"${This.FTLModifiers}${key.Escape}\""]
			}
			else
			{
				This:AddReleaseCommand["relay \"${This.ResolveTarget["${name.Escape}"]}\" press -nomodifiers \"${This.FTLModifiers}${key.Escape}\""]
			}
			
		}
	}

	method AddLightAction(string uplink, string name, string value)
	{
		if ${name.Equal[None]}
			return
		variable string command
		if ${uplink.Length} && ${uplink.NotEqual[NULL]}
			command:Set["relay \"${uplink}\" "]
		command:Concat["keymapper -light \"${name.Escape}\" \"${value.Escape}\""]

		This:AddReleaseCommand["${command.Escape}"]					
	}

	method AddKeyMapAction(string target, string name, string value)
	{
		variable string command
		if ${target.Length}
			command:Set["relay \"${This.ResolveTarget["${target}"]}\" "]

		switch ${value}
		{
		case Toggle
			command:Concat["keymapper -togglemap \"${name.Escape}\""]
			break
		case On
			command:Concat["keymapper -loadmap \"${name.Escape}\""]
			break
		case Off
			command:Concat["keymapper -unloadmap \"${name.Escape}\""]
			break
		}
		This:AddReleaseCommand["${command.Escape}"]									
	}


	method CreateAlias()
	{
		if ${Hold}
		{
			execute squelch alias "${Name.Escape}_press" "${PressCmd.Escape}"
		}
		execute squelch alias "${Name.Escape}_release" "${ReleaseCmd.Escape}"
	}

	method CreateBind()
	{
		if ${PressCmd.Length}
		{
			Input:Bind[-press,"${Name.Escape}","${Hotkey.Escape}","${PressCmd.Escape}"]
			
		}
		if ${ReleaseCmd.Length}
		{
			if ${SendNextClick.Length}
			{
				if ${SendNextClick.Equal[${Int[${SendNextClick}]}]}
					Input:Bind[-release,"${Name.Escape}","${Hotkey.Escape}","${ReleaseCmd.Escape};noop \${MouseRepeater:SendNextClick[\"is${SendNextClick}\",\"${Name.Escape}\"]}"]
				else
					Input:Bind[-release,"${Name.Escape}","${Hotkey.Escape}","${ReleaseCmd.Escape};noop \${MouseRepeater:SendNextClick[\"${SendNextClick}\",\"${Name.Escape}\"]}"]
			}
			else
			{
				Input:Bind[-release,"${Name.Escape}","${Hotkey.Escape}","${ReleaseCmd.Escape}"]
			}
		}
		else
		{
			if ${SendNextClick.Length}
			{
				if ${SendNextClick.Equal[${Int[${SendNextClick}]}]}
					Input:Bind[-release,"${Name.Escape}","${Hotkey.Escape}","noop \${MouseRepeater:SendNextClick[\"is${SendNextClick}\",\"${Name.Escape}\"]}"]
				else
					Input:Bind[-release,"${Name.Escape}","${Hotkey.Escape}","noop \${MouseRepeater:SendNextClick[\"${SendNextClick}\",\"${Name.Escape}\"]}"]
			}
		}
	}
}

objectdef keymapper
{
	variable settingsetref ListRoot
	variable string FTLModifiers
	variable set LoadedMaps

	method Initialize()
	{
		LavishSettings:AddSet[KeyMapper]
		ListRoot:Set[${LavishSettings.FindSet["KeyMapper"]}]
	}

	method Shutdown()
	{
		ListRoot:Remove
		squelch alias -delete keymapper_*
		squelch bind -delete keymapper_*
		echo Keymapper: Key maps unloaded.
	}

	method UpdateFTLModifiers()
	{
		variable settingsetref ftlSet
		variable int nSession=${Session.Right[-2]}

		ftlSet:Set[${ListRoot.FindSet["FTL Modifiers"]}]
		if !${ftlSet.Children(exists)}
			return
		
		if ${ftlSet.FindSetting[${nSession}](exists)}
		{
			This:SetFTLModifiers["${ftlSet.FindSetting[${nSession}].String.Escape}"]
		}
		else
		{
			FTLModifiers:Set[""]
		}
	}

	method SetFTLModifiers(string mods)
	{
		FTLModifiers:Set["${mods.Escape}"]
		if ${FTLModifiers.Length}
		{
			if ${FTLModifiers.Right[1].NotEqual["+"]}
				FTLModifiers:Concat["+"]
		}
		echo Keymapper: FTLModifiers=${FTLModifiers.Escape}
	}

	method Import(string filename)
	{
		echo Keymapper: Importing key maps from '\at${filename}\ay'
		if !${ListRoot:Import["${filename.Escape}"](exists)}
		{
			echo Keymapper: Reading of file '\at${filename}\ay' \arfailed\ax. It either doesn't exist at the given location, or is invalid XML.
			return
		}
		This:UpdateFTLModifiers
	}

	method ToggleKeyMap(string name)
	{
		if ${LoadedMaps.Contains["${name.Escape}"]}
			This:UnloadKeyMap["${name.Escape}"]
		else
			This:LoadKeyMap["${name.Escape}"]
	}

	method LoadKeyMap(string name)
	{
		echo Keymapper: Loading key map '\at${name.Escape}\ax'
		This:LoadKeyMapFromSet["${name.Escape}",${ListRoot.FindSet["key maps"].FindSet["${name.Escape}"]}]
	}

	method LoadAlias(string main, string name, settingsetref loadSet, bool hold, bool useftl, string nextStep)
	{
		variable bindbuilder bb
		bb.Name:Set["${name.Escape}"]
		bb.Hold:Set[${hold}]
		bb.UseFTLModifiers:Set[${useftl}]
		
		variable iterator Iterator
		loadSet:GetSettingIterator[Iterator]
		if ${Iterator:First(exists)}
		{
			do
			{
				if ${Iterator.Value.FindAttribute[ActionType](exists)}
				{
					switch ${Iterator.Value.FindAttribute[ActionType]}
					{
					case Light
						bb:AddLightAction["${Iterator.Value.FindAttribute[Computer].String.Escape}","${Iterator.Value.String.Escape}","${Iterator.Value.FindAttribute[Action].String.Escape}"]
						break
					case KeyMap
						bb:AddKeyMapAction["${Iterator.Value.FindAttribute[Target].String.Escape}","${Iterator.Value.String.Escape}","${Iterator.Value.FindAttribute[Action].String.Escape}"]
						break
					}
				}
				else
					bb:AddKeypress["${Iterator.Key.Escape}","${Iterator.Value.String.Escape}"]
			}
			while ${Iterator:Next(exists)}
		}

		if ${hold}
			bb:AddPressCommand["squelch alias \"${main.Escape}_press\" \"${nextStep.Escape}_press\""]
		bb:AddReleaseCommand["squelch alias \"${main.Escape}_release\" \"${nextStep.Escape}_release\""]

		bb:CreateAlias
	}

	method LoadBind(string name, settingsetref loadSet, bool useftl, bool hold)
	{
		if !${loadSet.FindSetting["Hotkey"](exists)}
		{
			echo Keymapper: Cannot load bind ${loadSet.Name}: no hotkey is set
			return
		}

		variable bindbuilder bb
		name:Set["${name.Replace[" ","_"].Escape}"]
		bb.Name:Set["${name.Escape}"]
		bb.Hotkey:Set["${loadSet.FindSetting["Hotkey"].String.Escape}"]

		if ${loadSet.FindSetting["Send Next Click"].String.Length}
			bb.SendNextClick:Set["${loadSet.FindSetting["Send Next Click"].String.Escape}"]

		if ${loadSet.FindSetting["Hold"].String.Length}
			bb.Hold:Set["${loadSet.FindSetting["Hold"].String.Escape}"]
		else
			bb.Hold:Set["${hold}"]

		if ${loadSet.FindSetting["Use FTL Modifiers"].String.Length}
			bb.UseFTLModifiers:Set["${loadSet.FindSetting["Use FTL Modifiers"].String.Escape}"]
		else
			bb.UseFTLModifiers:Set["${useftl}"]

		variable iterator Iterator
		loadSet:GetSettingIterator[Iterator]
		if ${Iterator:First(exists)}
		{
			do
			{
				switch ${Iterator.Key}
				{
				case Hold
				case Use FTL Modifiers
				case Hotkey
				case Send Next Click
					break
				default
					if ${Iterator.Value.FindAttribute[ActionType](exists)}
					{
						switch ${Iterator.Value.FindAttribute[ActionType]}
						{
						case Light
							bb:AddLightAction["${Iterator.Value.FindAttribute[Computer].String.Escape}","${Iterator.Value.String.Escape}","${Iterator.Value.FindAttribute[Action].String.Escape}"]
							break
						case KeyMap
							bb:AddKeyMapAction["${Iterator.Value.FindAttribute[Target].String.Escape}","${Iterator.Value.String.Escape}","${Iterator.Value.FindAttribute[Action].String.Escape}"]
							break
						}
					}
					else
						bb:AddKeypress["${Iterator.Key.Escape}","${Iterator.Value.String.Escape}"]
				}
			}
			while ${Iterator:Next(exists)}
		}
		loadSet:GetSetIterator[Iterator]
		if ${Iterator:First(exists)}
		{
			if ${bb.Hold}
				execute squelch alias "${name.Escape}_press" "${name.Escape}_${Iterator.Key.Replace[" ","_"].Escape}_press"
			execute squelch alias "${name.Escape}_release" "${name.Escape}_${Iterator.Key.Replace[" ","_"].Escape}_release"
			variable settingsetref firstStep
			variable settingsetref thisStep
			firstStep:Set[${Iterator.Value}]
			while 1
			{
				thisStep:Set[${Iterator.Value}]
				if ${Iterator:Next(exists)}
				{
					This:LoadAlias["${name.Escape}","${name.Escape}_${thisStep.Name.Replace[" ","_"].Escape}",${thisStep},${bb.Hold},${bb.UseFTLModifiers},"${name.Escape}_${Iterator.Key.Replace[" ","_"].Escape}"]
				}
				else
				{
					This:LoadAlias["${name.Escape}","${name.Escape}_${thisStep.Name.Replace[" ","_"].Escape}",${thisStep},${bb.Hold},${bb.UseFTLModifiers},"${name.Escape}_${firstStep.Name.Replace[" ","_"].Escape}"]
					break
				}
			}

			if ${bb.Hold}
				bb:AddPressCommand["${name.Escape}_press"]
			bb:AddReleaseCommand["${name.Escape}_release"]
		}

		bb:CreateBind
	}

	method LoadKeyMapFromSet(string name,settingsetref loadSet)
	{
		variable bool useftl
		useftl:Set["${loadSet.FindSetting["Use FTL Modifiers"].String.Escape}"]
		variable bool hold
		hold:Set["${loadSet.FindSetting["Hold"].String.Escape}"]

		variable iterator Iterator
		loadSet:GetSetIterator[Iterator]

		if ${Iterator:First(exists)}
		{
			do
			{
				This:LoadBind["keymapper_${loadSet.Name.Escape}_${Iterator.Key.Escape}","${Iterator.Value}","${useftl}","${hold}"]
			}
			while ${Iterator:Next(exists)}
			LoadedMaps:Add["${name.Escape}"]		
		}
		else
		{
			echo Keymapper: Key map is empty or not found
		}						
	}

	method UnloadKeyMap(string name)
	{
		execute squelch bind -delete "keymapper_${name.Escape}*"
		execute squelch alias -delete "keymapper_${name.Escape}*"
		LoadedMaps:Remove["${name.Escape}"]		
	}

	method LightAction(string light, string value)
	{
		switch ${light}
		{
		case M1
		case M2
		case M3
		case MR
			switch ${value}
			{
			case On
				Input.G15:Set${light.Upper}Light[TRUE]
				break
			case Off
				Input.G15:Set${light.Upper}Light[FALSE]
				break
			case Toggle
				Input.G15:Set${light.Upper}Light[!${Input.G15.${light.Upper}Light}]				
				break
			}
			break
		}
	}

}

variable keymapper KeyMapper

atom(global) keymapper(... Params)
{
	/* Process parameters */
	variable int p=1
	for (${p}<=${Params.Size} ; p:Inc)
	{
		/* Switch does a string comparison, so parameter processing is easy */
		switch ${Params[${p}]}
		{
		case -import
			p:Inc
			if ${p}<=${Params.Size}
			{
				KeyMapper:Import["${Params[${p}].Escape}"]
			}
			break
		case -togglemap
			p:Inc
			if ${p}<=${Params.Size}
			{
				KeyMapper:ToggleKeyMap["${Params[${p}].Escape}"]
			}
			break
		case -loadmap
			p:Inc
			if ${p}<=${Params.Size}
			{
				KeyMapper:LoadKeyMap["${Params[${p}].Escape}"]
			}
			break
		case -unloadmap
			p:Inc
			if ${p}<=${Params.Size}
			{
				KeyMapper:UnloadKeyMap["${Params[${p}].Escape}"]
			}
			break
		case -ftlmods
			p:Inc
			if ${p}<=${Params.Size}
			{			
				KeyMapper:SetFTLModifiers["${Params[${p}].Escape}"]
			}
			break
		case -noftlmods
			KeyMapper.FTLModifiers:Set[""]
			break
		case -light
			p:Inc
			if ${p}<=${Params.Size}
			{			
				KeyMapper:LightAction["${Params[${p}].Escape}","${Params[${p:Inc}].Escape}"]
			}
			break
		}
	}			
}

function main(... Params)
{
	KeyMapper:Import[keymapper.xml]

	KeyMapper:LoadKeyMap[main]


	while 1
		waitframe
}

